"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.resolve = exports.relative = exports.extname = exports.dirname = exports.basename = exports.normalize = exports.isRelative = exports.isAbsolute = exports.relativePath = exports.absolutePath = void 0;
exports.absolutePath = /^(?:\/|(?:[A-Za-z]:)?[\\|/])/;
exports.relativePath = /^\.?\.\//;
function isAbsolute(path) {
    return exports.absolutePath.test(path);
}
exports.isAbsolute = isAbsolute;
function isRelative(path) {
    return exports.relativePath.test(path);
}
exports.isRelative = isRelative;
function normalize(path) {
    return path.replace(/\\/g, '/');
}
exports.normalize = normalize;
function basename(path) {
    return path.split(/(\/|\\)/).pop();
}
exports.basename = basename;
function dirname(path) {
    const match = /(\/|\\)[^/\\]*$/.exec(path);
    if (!match)
        return '.';
    const dir = path.slice(0, -match[0].length);
    // If `dir` is the empty string, we're at root.
    return dir ? dir : '/';
}
exports.dirname = dirname;
function extname(path) {
    const match = /\.[^.]+$/.exec(basename(path));
    if (!match)
        return '';
    return match[0];
}
exports.extname = extname;
function relative(from, to) {
    const fromParts = from.split(/[/\\]/).filter(Boolean);
    const toParts = to.split(/[/\\]/).filter(Boolean);
    if (fromParts[0] === '.')
        fromParts.shift();
    if (toParts[0] === '.')
        toParts.shift();
    while (fromParts[0] && toParts[0] && fromParts[0] === toParts[0]) {
        fromParts.shift();
        toParts.shift();
    }
    while (toParts[0] === '..' && fromParts.length > 0) {
        toParts.shift();
        fromParts.pop();
    }
    while (fromParts.pop()) {
        toParts.unshift('..');
    }
    return toParts.join('/');
}
exports.relative = relative;
function resolve(...paths) {
    let resolvedParts = paths.shift().split(/[/\\]/);
    paths.forEach(path => {
        if (isAbsolute(path)) {
            resolvedParts = path.split(/[/\\]/);
        }
        else {
            const parts = path.split(/[/\\]/);
            while (parts[0] === '.' || parts[0] === '..') {
                const part = parts.shift();
                if (part === '..') {
                    resolvedParts.pop();
                }
            }
            resolvedParts.push.apply(resolvedParts, parts);
        }
    });
    return resolvedParts.join('/');
}
exports.resolve = resolve;
