"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.handleError = exports.stderr = void 0;
const turbocolor_1 = __importDefault(require("turbocolor"));
const relativeId_1 = __importDefault(require("../src/utils/relativeId"));
// @see https://no-color.org
// @see https://www.npmjs.com/package/chalk
if (process.env.FORCE_COLOR === '0' || process.env.NO_COLOR) {
    turbocolor_1.default.enabled = false;
}
// log to stderr to keep `rollup main.js > bundle.js` from breaking
exports.stderr = console.error.bind(console);
function handleError(err, recover = false) {
    let description = err.message || err;
    if (err.name)
        description = `${err.name}: ${description}`;
    const message = (err.plugin
        ? `(plugin ${(err).plugin}) ${description}`
        : description) || err;
    exports.stderr(turbocolor_1.default.bold.red(`[!] ${turbocolor_1.default.bold(message.toString())}`));
    if (err.url) {
        exports.stderr(turbocolor_1.default.cyan(err.url));
    }
    if (err.loc) {
        exports.stderr(`${relativeId_1.default((err.loc.file || err.id))} (${err.loc.line}:${err.loc.column})`);
    }
    else if (err.id) {
        exports.stderr(relativeId_1.default(err.id));
    }
    if (err.frame) {
        exports.stderr(turbocolor_1.default.dim(err.frame));
    }
    if (err.stack) {
        exports.stderr(turbocolor_1.default.dim(err.stack));
    }
    exports.stderr('');
    if (!recover)
        process.exit(1);
}
exports.handleError = handleError;
