"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const path = __importStar(require("path"));
const require_relative_1 = __importDefault(require("require-relative"));
const mergeOptions_1 = __importDefault(require("../../src/utils/mergeOptions"));
const relativeId_1 = require("../../src/utils/relativeId");
const logging_1 = require("../logging");
const batchWarnings_1 = __importDefault(require("./batchWarnings"));
const build_1 = __importDefault(require("./build"));
const loadConfigFile_1 = __importDefault(require("./loadConfigFile"));
const stdin_1 = require("./stdin");
const watch_1 = __importDefault(require("./watch"));
function runRollup(command) {
    let inputSource;
    if (command._.length > 0) {
        if (command.input) {
            logging_1.handleError({
                code: 'DUPLICATE_IMPORT_OPTIONS',
                message: 'use --input, or pass input path as argument'
            });
        }
        inputSource = command._;
    }
    else if (typeof command.input === 'string') {
        inputSource = [command.input];
    }
    else {
        inputSource = command.input;
    }
    if (inputSource && inputSource.length > 0) {
        if (inputSource.some((input) => input.indexOf('=') !== -1)) {
            command.input = {};
            inputSource.forEach((input) => {
                const equalsIndex = input.indexOf('=');
                const value = input.substr(equalsIndex + 1);
                let key = input.substr(0, equalsIndex);
                if (!key)
                    key = relativeId_1.getAliasName(input);
                command.input[key] = value;
            });
        }
        else {
            command.input = inputSource;
        }
    }
    if (command.environment) {
        const environment = Array.isArray(command.environment)
            ? command.environment
            : [command.environment];
        environment.forEach((arg) => {
            arg.split(',').forEach((pair) => {
                const [key, ...value] = pair.split(':');
                if (value.length) {
                    process.env[key] = value.join(':');
                }
                else {
                    process.env[key] = String(true);
                }
            });
        });
    }
    let configFile = command.config === true ? 'rollup.config.js' : command.config;
    if (configFile) {
        if (configFile.slice(0, 5) === 'node:') {
            const pkgName = configFile.slice(5);
            try {
                configFile = require_relative_1.default.resolve(`rollup-config-${pkgName}`, process.cwd());
            }
            catch (err) {
                try {
                    configFile = require_relative_1.default.resolve(pkgName, process.cwd());
                }
                catch (err) {
                    if (err.code === 'MODULE_NOT_FOUND') {
                        logging_1.handleError({
                            code: 'MISSING_EXTERNAL_CONFIG',
                            message: `Could not resolve config file ${configFile}`
                        });
                    }
                    throw err;
                }
            }
        }
        else {
            // find real path of config so it matches what Node provides to callbacks in require.extensions
            configFile = fs_1.realpathSync(configFile);
        }
        if (command.watch)
            process.env.ROLLUP_WATCH = 'true';
        return loadConfigFile_1.default(configFile, command)
            .then(configs => execute(configFile, configs, command))
            .catch(logging_1.handleError);
    }
    else {
        if (!command.input && (command.stdin || !process.stdin.isTTY)) {
            command.input = stdin_1.stdinName;
        }
        return execute(configFile, [{ input: [] }], command);
    }
}
exports.default = runRollup;
function execute(configFile, configs, command) {
    return __awaiter(this, void 0, void 0, function* () {
        if (command.watch) {
            watch_1.default(configFile, configs, command, command.silent);
        }
        else {
            for (const config of configs) {
                const warnings = batchWarnings_1.default();
                try {
                    const { inputOptions, outputOptions, optionError } = mergeOptions_1.default({
                        command,
                        config,
                        defaultOnWarnHandler: warnings.add
                    });
                    if (optionError) {
                        inputOptions.onwarn({ code: 'UNKNOWN_OPTION', message: optionError });
                    }
                    if (command.stdin !== false) {
                        inputOptions.plugins.push(stdin_1.stdinPlugin());
                    }
                    if (command.plugin) {
                        const plugins = Array.isArray(command.plugin) ? command.plugin : [command.plugin];
                        for (const plugin of plugins) {
                            if (/[={}]/.test(plugin)) {
                                // -p plugin=value
                                // -p "{transform(c,i){...}}"
                                loadAndRegisterPlugin(inputOptions, plugin);
                            }
                            else {
                                // split out plugins joined by commas
                                // -p node-resolve,commonjs,buble
                                plugin
                                    .split(',')
                                    .forEach((plugin) => loadAndRegisterPlugin(inputOptions, plugin));
                            }
                        }
                    }
                    yield build_1.default(inputOptions, outputOptions, warnings, command.silent);
                }
                catch (err) {
                    warnings.flush();
                    logging_1.handleError(err);
                }
            }
        }
    });
}
function loadAndRegisterPlugin(inputOptions, pluginText) {
    let plugin = null;
    let pluginArg = undefined;
    if (pluginText[0] === '{') {
        // -p "{transform(c,i){...}}"
        plugin = new Function('return ' + pluginText);
    }
    else {
        const match = pluginText.match(/^([@.\/\\\w|^{}|-]+)(=(.*))?$/);
        if (match) {
            // -p plugin
            // -p plugin=arg
            pluginText = match[1];
            pluginArg = new Function('return ' + match[3])();
        }
        else {
            throw new Error(`Invalid --plugin argument format: ${JSON.stringify(pluginText)}`);
        }
        if (!/^\.|^rollup-plugin-|[@\/\\]/.test(pluginText)) {
            // Try using plugin prefix variations first if applicable.
            // Prefix order is significant - left has higher precedence.
            for (const prefix of ['@rollup/plugin-', 'rollup-plugin-']) {
                try {
                    plugin = require(prefix + pluginText);
                    break;
                }
                catch (ex) {
                    // if this does not work, we try requiring the actual name below
                }
            }
        }
        if (!plugin) {
            try {
                if (pluginText[0] == '.')
                    pluginText = path.resolve(pluginText);
                plugin = require(pluginText);
            }
            catch (ex) {
                throw new Error(`Cannot load plugin "${pluginText}"`);
            }
        }
    }
    if (typeof plugin === 'object' && pluginText in plugin) {
        // some plugins do not use `export default` for their entry point.
        // attempt to use the plugin name as the named import name.
        plugin = plugin[pluginText];
    }
    inputOptions.plugins.push(typeof plugin === 'function' ? plugin.call(plugin, pluginArg) : plugin);
}
