"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const PathTracker_1 = require("../utils/PathTracker");
const values_1 = require("../values");
const Node_1 = require("./shared/Node");
class ArrayPattern extends Node_1.NodeBase {
    addExportedVariables(variables) {
        for (const element of this.elements) {
            if (element !== null) {
                element.addExportedVariables(variables);
            }
        }
    }
    declare(kind) {
        const variables = [];
        for (const element of this.elements) {
            if (element !== null) {
                variables.push(...element.declare(kind, values_1.UNKNOWN_EXPRESSION));
            }
        }
        return variables;
    }
    deoptimizePath(path) {
        if (path.length === 0) {
            for (const element of this.elements) {
                if (element !== null) {
                    element.deoptimizePath(path);
                }
            }
        }
    }
    hasEffectsWhenAssignedAtPath(path, context) {
        if (path.length > 0)
            return true;
        for (const element of this.elements) {
            if (element !== null && element.hasEffectsWhenAssignedAtPath(PathTracker_1.EMPTY_PATH, context))
                return true;
        }
        return false;
    }
}
exports.default = ArrayPattern;
