"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const PathTracker_1 = require("../utils/PathTracker");
const values_1 = require("../values");
const ExpressionStatement_1 = __importDefault(require("./ExpressionStatement"));
const Node_1 = require("./shared/Node");
const binaryOperators = {
    '!=': (left, right) => left != right,
    '!==': (left, right) => left !== right,
    '%': (left, right) => left % right,
    '&': (left, right) => left & right,
    '*': (left, right) => left * right,
    // At the moment, "**" will be transpiled to Math.pow
    '**': (left, right) => Math.pow(left, right),
    '+': (left, right) => left + right,
    '-': (left, right) => left - right,
    '/': (left, right) => left / right,
    '<': (left, right) => left < right,
    '<<': (left, right) => left << right,
    '<=': (left, right) => left <= right,
    '==': (left, right) => left == right,
    '===': (left, right) => left === right,
    '>': (left, right) => left > right,
    '>=': (left, right) => left >= right,
    '>>': (left, right) => left >> right,
    '>>>': (left, right) => left >>> right,
    '^': (left, right) => left ^ right,
    in: () => values_1.UnknownValue,
    instanceof: () => values_1.UnknownValue,
    '|': (left, right) => left | right
};
class BinaryExpression extends Node_1.NodeBase {
    deoptimizeCache() { }
    getLiteralValueAtPath(path, recursionTracker, origin) {
        if (path.length > 0)
            return values_1.UnknownValue;
        const leftValue = this.left.getLiteralValueAtPath(PathTracker_1.EMPTY_PATH, recursionTracker, origin);
        if (leftValue === values_1.UnknownValue)
            return values_1.UnknownValue;
        const rightValue = this.right.getLiteralValueAtPath(PathTracker_1.EMPTY_PATH, recursionTracker, origin);
        if (rightValue === values_1.UnknownValue)
            return values_1.UnknownValue;
        const operatorFn = binaryOperators[this.operator];
        if (!operatorFn)
            return values_1.UnknownValue;
        return operatorFn(leftValue, rightValue);
    }
    hasEffects(context) {
        // support some implicit type coercion runtime errors
        if (this.operator === '+' &&
            this.parent instanceof ExpressionStatement_1.default &&
            this.left.getLiteralValueAtPath(PathTracker_1.EMPTY_PATH, PathTracker_1.SHARED_RECURSION_TRACKER, this) === '')
            return true;
        return super.hasEffects(context);
    }
    hasEffectsWhenAccessedAtPath(path) {
        return path.length > 1;
    }
}
exports.default = BinaryExpression;
