"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const renderHelpers_1 = require("../../utils/renderHelpers");
const Node_1 = require("./shared/Node");
class SwitchCase extends Node_1.NodeBase {
    hasEffects(context) {
        if (this.test && this.test.hasEffects(context))
            return true;
        for (const node of this.consequent) {
            if (context.brokenFlow)
                break;
            if (node.hasEffects(context))
                return true;
        }
        return false;
    }
    include(context, includeChildrenRecursively) {
        this.included = true;
        if (this.test)
            this.test.include(context, includeChildrenRecursively);
        for (const node of this.consequent) {
            if (includeChildrenRecursively || node.shouldBeIncluded(context))
                node.include(context, includeChildrenRecursively);
        }
    }
    render(code, options, nodeRenderOptions) {
        if (this.consequent.length) {
            this.test && this.test.render(code, options);
            const testEnd = this.test
                ? this.test.end
                : renderHelpers_1.findFirstOccurrenceOutsideComment(code.original, 'default', this.start) + 7;
            const consequentStart = renderHelpers_1.findFirstOccurrenceOutsideComment(code.original, ':', testEnd) + 1;
            renderHelpers_1.renderStatementList(this.consequent, code, consequentStart, nodeRenderOptions.end, options);
        }
        else {
            super.render(code, options);
        }
    }
}
exports.default = SwitchCase;
SwitchCase.prototype.needsBoundaries = true;
