"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const values_1 = require("../values");
const ExportDefaultVariable_1 = __importDefault(require("../variables/ExportDefaultVariable"));
const GlobalVariable_1 = __importDefault(require("../variables/GlobalVariable"));
const LocalVariable_1 = __importDefault(require("../variables/LocalVariable"));
const ChildScope_1 = __importDefault(require("./ChildScope"));
class ModuleScope extends ChildScope_1.default {
    constructor(parent, context) {
        super(parent);
        this.context = context;
        this.variables.set('this', new LocalVariable_1.default('this', null, values_1.UNDEFINED_EXPRESSION, context));
    }
    addExportDefaultDeclaration(name, exportDefaultDeclaration, context) {
        const variable = new ExportDefaultVariable_1.default(name, exportDefaultDeclaration, context);
        this.variables.set('default', variable);
        return variable;
    }
    addNamespaceMemberAccess(_name, variable) {
        if (variable instanceof GlobalVariable_1.default) {
            this.accessedOutsideVariables.set(variable.name, variable);
        }
    }
    deconflict(format) {
        // all module level variables are already deconflicted when deconflicting the chunk
        for (const scope of this.children)
            scope.deconflict(format);
    }
    findLexicalBoundary() {
        return this;
    }
    findVariable(name) {
        const knownVariable = this.variables.get(name) || this.accessedOutsideVariables.get(name);
        if (knownVariable) {
            return knownVariable;
        }
        const variable = this.context.traceVariable(name) || this.parent.findVariable(name);
        if (variable instanceof GlobalVariable_1.default) {
            this.accessedOutsideVariables.set(name, variable);
        }
        return variable;
    }
}
exports.default = ModuleScope;
