"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.assignChunkColouringHashes = void 0;
const ExternalModule_1 = __importDefault(require("../ExternalModule"));
const Module_1 = __importDefault(require("../Module"));
const entryHashing_1 = require("./entryHashing");
function assignChunkColouringHashes(entryModules, manualChunkModules) {
    const { dependentEntryPointsByModule, dynamicImportersByModule } = analyzeModuleGraph(entryModules);
    const dynamicDependentEntryPointsByDynamicEntry = getDynamicDependentEntryPoints(dependentEntryPointsByModule, dynamicImportersByModule);
    const staticEntries = new Set(entryModules);
    function addColourToModuleDependencies(entry, colour, dynamicDependentEntryPoints) {
        const manualChunkAlias = entry.manualChunkAlias;
        const modulesToHandle = new Set([entry]);
        for (const module of modulesToHandle) {
            if (manualChunkAlias) {
                module.manualChunkAlias = manualChunkAlias;
                module.entryPointsHash = colour;
            }
            else if (dynamicDependentEntryPoints &&
                areEntryPointsContainedOrDynamicallyDependent(dynamicDependentEntryPoints, dependentEntryPointsByModule.get(module))) {
                continue;
            }
            else {
                entryHashing_1.Uint8ArrayXor(module.entryPointsHash, colour);
            }
            for (const dependency of module.dependencies) {
                if (!(dependency instanceof ExternalModule_1.default || dependency.manualChunkAlias)) {
                    modulesToHandle.add(dependency);
                }
            }
        }
    }
    function areEntryPointsContainedOrDynamicallyDependent(entryPoints, superSet) {
        const entriesToCheck = new Set(entryPoints);
        for (const entry of entriesToCheck) {
            if (!superSet.has(entry)) {
                if (staticEntries.has(entry))
                    return false;
                const dynamicDependentEntryPoints = dynamicDependentEntryPointsByDynamicEntry.get(entry);
                for (const dependentEntry of dynamicDependentEntryPoints) {
                    entriesToCheck.add(dependentEntry);
                }
            }
        }
        return true;
    }
    if (manualChunkModules) {
        for (const chunkName of Object.keys(manualChunkModules)) {
            const entryHash = entryHashing_1.randomUint8Array(10);
            for (const entry of manualChunkModules[chunkName]) {
                addColourToModuleDependencies(entry, entryHash, null);
            }
        }
    }
    for (const entry of entryModules) {
        if (!entry.manualChunkAlias) {
            const entryHash = entryHashing_1.randomUint8Array(10);
            addColourToModuleDependencies(entry, entryHash, null);
        }
    }
    for (const entry of dynamicImportersByModule.keys()) {
        if (!entry.manualChunkAlias) {
            const entryHash = entryHashing_1.randomUint8Array(10);
            addColourToModuleDependencies(entry, entryHash, dynamicDependentEntryPointsByDynamicEntry.get(entry));
        }
    }
}
exports.assignChunkColouringHashes = assignChunkColouringHashes;
function analyzeModuleGraph(entryModules) {
    const dynamicImportersByModule = new Map();
    const dependentEntryPointsByModule = new Map();
    const entriesToHandle = new Set(entryModules);
    for (const currentEntry of entriesToHandle) {
        const modulesToHandle = new Set([currentEntry]);
        for (const module of modulesToHandle) {
            getDependentModules(dependentEntryPointsByModule, module).add(currentEntry);
            for (const dependency of module.dependencies) {
                if (!(dependency instanceof ExternalModule_1.default)) {
                    modulesToHandle.add(dependency);
                }
            }
            for (const { resolution } of module.dynamicImports) {
                if (resolution instanceof Module_1.default &&
                    resolution.dynamicallyImportedBy.length > 0 &&
                    !resolution.manualChunkAlias) {
                    getDependentModules(dynamicImportersByModule, resolution).add(module);
                    entriesToHandle.add(resolution);
                }
            }
        }
    }
    return { dependentEntryPointsByModule, dynamicImportersByModule };
}
function getDependentModules(moduleMap, module) {
    const dependentModules = moduleMap.get(module) || new Set();
    moduleMap.set(module, dependentModules);
    return dependentModules;
}
function getDynamicDependentEntryPoints(dependentEntryPointsByModule, dynamicImportersByModule) {
    const dynamicDependentEntryPointsByDynamicEntry = new Map();
    for (const [dynamicEntry, importers] of dynamicImportersByModule.entries()) {
        const dynamicDependentEntryPoints = getDependentModules(dynamicDependentEntryPointsByDynamicEntry, dynamicEntry);
        for (const importer of importers) {
            for (const entryPoint of dependentEntryPointsByModule.get(importer)) {
                dynamicDependentEntryPoints.add(entryPoint);
            }
        }
    }
    return dynamicDependentEntryPointsByDynamicEntry;
}
