"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.warnDeprecatedHooks = exports.deprecatedHooks = exports.throwPluginError = exports.ANONYMOUS_OUTPUT_PLUGIN_PREFIX = exports.ANONYMOUS_PLUGIN_PREFIX = void 0;
const error_1 = require("./error");
exports.ANONYMOUS_PLUGIN_PREFIX = 'at position ';
exports.ANONYMOUS_OUTPUT_PLUGIN_PREFIX = 'at output position ';
function throwPluginError(err, plugin, { hook, id } = {}) {
    if (typeof err === 'string')
        err = { message: err };
    if (err.code && err.code !== error_1.Errors.PLUGIN_ERROR) {
        err.pluginCode = err.code;
    }
    err.code = error_1.Errors.PLUGIN_ERROR;
    err.plugin = plugin;
    if (hook) {
        err.hook = hook;
    }
    if (id) {
        err.id = id;
    }
    return error_1.error(err);
}
exports.throwPluginError = throwPluginError;
exports.deprecatedHooks = [
    { active: true, deprecated: 'ongenerate', replacement: 'generateBundle' },
    { active: true, deprecated: 'onwrite', replacement: 'generateBundle/writeBundle' },
    { active: true, deprecated: 'transformBundle', replacement: 'renderChunk' },
    { active: true, deprecated: 'transformChunk', replacement: 'renderChunk' },
    { active: false, deprecated: 'resolveAssetUrl', replacement: 'resolveFileUrl' }
];
function warnDeprecatedHooks(plugins, graph) {
    for (const { active, deprecated, replacement } of exports.deprecatedHooks) {
        for (const plugin of plugins) {
            if (deprecated in plugin) {
                graph.warnDeprecation({
                    message: `The "${deprecated}" hook used by plugin ${plugin.name} is deprecated. The "${replacement}" hook should be used instead.`,
                    plugin: plugin.name
                }, active);
            }
        }
    }
}
exports.warnDeprecatedHooks = warnDeprecatedHooks;
