"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeUnique = exports.renderNamePattern = void 0;
const error_1 = require("./error");
const path_1 = require("./path");
const relativeId_1 = require("./relativeId");
function renderNamePattern(pattern, patternName, replacements) {
    if (!relativeId_1.isPlainPathFragment(pattern))
        return error_1.error(error_1.errFailedValidation(`Invalid pattern "${pattern}" for "${patternName}", patterns can be neither absolute nor relative paths and must not contain invalid characters.`));
    return pattern.replace(/\[(\w+)\]/g, (_match, type) => {
        if (!replacements.hasOwnProperty(type)) {
            return error_1.error(error_1.errFailedValidation(`"[${type}]" is not a valid placeholder in "${patternName}" pattern.`));
        }
        const replacement = replacements[type]();
        if (!relativeId_1.isPlainPathFragment(replacement))
            return error_1.error(error_1.errFailedValidation(`Invalid substitution "${replacement}" for placeholder "[${type}]" in "${patternName}" pattern, can be neither absolute nor relative path.`));
        return replacement;
    });
}
exports.renderNamePattern = renderNamePattern;
function makeUnique(name, existingNames) {
    const existingNamesLowercase = new Set(Object.keys(existingNames).map(key => key.toLowerCase()));
    if (!existingNamesLowercase.has(name.toLocaleLowerCase()))
        return name;
    const ext = path_1.extname(name);
    name = name.substr(0, name.length - ext.length);
    let uniqueName, uniqueIndex = 1;
    while (existingNamesLowercase.has((uniqueName = name + ++uniqueIndex + ext).toLowerCase()))
        ;
    return uniqueName;
}
exports.makeUnique = makeUnique;
